@@ -311,7 +311,7 @@ class XCodeBackend(backends.Backend):
         for fname, idval in self.filemap.items():
             fullpath = os.path.join(self.environment.get_source_dir(), fname)
             xcodetype = self.get_xcodetype(fname)
-            name = os.path.split(fname)[-1]
+            name = os.path.basename(fname)
             path = fname
             self.ofile.write(src_templ % (idval, fullpath, xcodetype, name, path))
         target_templ = '%s /* %s */ = { isa = PBXFileReference; explicitFileType = "%s"; path = %s; refType = %d; sourceTree = BUILT_PRODUCTS_DIR; };\n'
@@ -483,7 +483,6 @@ class XCodeBackend(backends.Backend):
                             % (self.buildconflistmap[tname], tname))
             self.write_line('buildPhases = (')
             self.indent_level += 1
-            t = self.build.targets[tname]
             for bpname, bpval in t.buildphasemap.items():
                 self.write_line('%s /* %s yyy */,' % (bpval, bpname))
             self.indent_level -= 1
@@ -568,7 +567,7 @@ class XCodeBackend(backends.Backend):
         self.write_line('shellPath = /bin/sh;')
         script_root = self.environment.get_script_dir()
         test_script = os.path.join(script_root, 'meson_test.py')
-        cmd = [sys.executable, test_script, test_data, '--wd', self.environment.get_build_dir()]
+        cmd = mesonlib.python_command + [test_script, test_data, '--wd', self.environment.get_build_dir()]
         cmdstr = ' '.join(["'%s'" % i for i in cmd])
         self.write_line('shellScript = "%s";' % cmdstr)
         self.write_line('showEnvVarsInLog = 0;')
