@@ -42,11 +42,11 @@ from .compilers import (
 
 
 class CCompiler(Compiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
         # If a child ObjC or CPP class has already set it, don't set it ourselves
         if not hasattr(self, 'language'):
             self.language = 'c'
-        super().__init__(exelist, version)
+        super().__init__(exelist, version, **kwargs)
         self.id = 'unknown'
         self.is_cross = is_cross
         self.can_compile_suffixes.add('h')
@@ -97,7 +97,7 @@ class CCompiler(Compiler):
         return self.build_unix_rpath_args(build_dir, from_dir, rpath_paths, build_rpath, install_rpath)
 
     def get_dependency_gen_args(self, outtarget, outfile):
-        return ['-MMD', '-MQ', outtarget, '-MF', outfile]
+        return ['-MD', '-MQ', outtarget, '-MF', outfile]
 
     def depfile_for_object(self, objfile):
         return objfile + '.' + self.get_depfile_suffix()
@@ -156,7 +156,9 @@ class CCompiler(Compiler):
         return ['-shared']
 
     def get_library_dirs(self):
-        stdo = Popen_safe(self.exelist + ['--print-search-dirs'])[1]
+        env = os.environ.copy()
+        env['LC_ALL'] = 'C'
+        stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=env)[1]
         for line in stdo.split('\n'):
             if line.startswith('libraries:'):
                 libstr = line.split('=', 1)[1]
@@ -170,10 +172,10 @@ class CCompiler(Compiler):
         return ' '.join(self.exelist)
 
     def get_pch_use_args(self, pch_dir, header):
-        return ['-include', os.path.split(header)[-1]]
+        return ['-include', os.path.basename(header)]
 
     def get_pch_name(self, header_name):
-        return os.path.split(header_name)[-1] + '.' + self.get_pch_suffix()
+        return os.path.basename(header_name) + '.' + self.get_pch_suffix()
 
     def get_linker_search_args(self, dirname):
         return ['-L' + dirname]
@@ -181,6 +183,14 @@ class CCompiler(Compiler):
     def get_default_include_dirs(self):
         return []
 
+    def gen_export_dynamic_link_args(self, env):
+        if for_windows(env.is_cross_build(), env) or for_cygwin(env.is_cross_build(), env):
+            return ['-Wl,--export-all-symbols']
+        elif for_darwin(env.is_cross_build(), env):
+            return []
+        else:
+            return ['-Wl,-export-dynamic']
+
     def gen_import_library_args(self, implibname):
         """
         The name of the outputted import library
@@ -797,12 +807,19 @@ class CCompiler(Compiler):
         return ['-pthread']
 
     def has_multi_arguments(self, args, env):
+        for arg in args:
+            if arg.startswith('-Wl,'):
+                mlog.warning('''{} looks like a linker argument, but has_argument
+and other similar methods only support checking compiler arguments.
+Using them to check linker arguments are never supported, and results
+are likely to be wrong regardless of the compiler you are using.
+'''.format(arg))
         return self.compiles('int i;\n', env, extra_args=args)
 
 
 class ClangCCompiler(ClangCompiler, CCompiler):
-    def __init__(self, exelist, version, clang_type, is_cross, exe_wrapper=None):
-        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
+    def __init__(self, exelist, version, clang_type, is_cross, exe_wrapper=None, **kwargs):
+        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
         ClangCompiler.__init__(self, clang_type)
         default_warn_args = ['-Wall', '-Winvalid-pch']
         self.warn_args = {'1': default_warn_args,
@@ -833,8 +850,8 @@ class ClangCCompiler(ClangCompiler, CCompiler):
 
 
 class GnuCCompiler(GnuCompiler, CCompiler):
-    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None):
-        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
+    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
+        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
         GnuCompiler.__init__(self, gcc_type, defines)
         default_warn_args = ['-Wall', '-Winvalid-pch']
         self.warn_args = {'1': default_warn_args,
@@ -868,12 +885,12 @@ class GnuCCompiler(GnuCompiler, CCompiler):
         return ['-shared']
 
     def get_pch_use_args(self, pch_dir, header):
-        return ['-fpch-preprocess', '-include', os.path.split(header)[-1]]
+        return ['-fpch-preprocess', '-include', os.path.basename(header)]
 
 
 class IntelCCompiler(IntelCompiler, CCompiler):
-    def __init__(self, exelist, version, icc_type, is_cross, exe_wrapper=None):
-        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
+    def __init__(self, exelist, version, icc_type, is_cross, exe_wrapper=None, **kwargs):
+        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
         IntelCompiler.__init__(self, icc_type)
         self.lang_header = 'c-header'
         default_warn_args = ['-Wall', '-w3', '-diag-disable:remark', '-Wpch-messages']
@@ -947,13 +964,13 @@ class VisualStudioCCompiler(CCompiler):
         return 'pch'
 
     def get_pch_name(self, header):
-        chopped = os.path.split(header)[-1].split('.')[:-1]
+        chopped = os.path.basename(header).split('.')[:-1]
         chopped.append(self.get_pch_suffix())
         pchname = '.'.join(chopped)
         return pchname
 
     def get_pch_use_args(self, pch_dir, header):
-        base = os.path.split(header)[-1]
+        base = os.path.basename(header)
         pchname = self.get_pch_name(header)
         return ['/FI' + base, '/Yu' + base, '/Fp' + os.path.join(pch_dir, pchname)]
 
@@ -1080,7 +1097,7 @@ class VisualStudioCCompiler(CCompiler):
         mlog.debug('Running VS compile:')
         mlog.debug('Command line: ', ' '.join(commands))
         mlog.debug('Code:\n', code)
-        p, stdo, stde = Popen_safe(commands, cwd=os.path.split(srcname)[0])
+        p, stdo, stde = Popen_safe(commands, cwd=os.path.dirname(srcname))
         if p.returncode != 0:
             return False
         return not(warning_text in stde or warning_text in stdo)
