@@ -1,4 +1,4 @@
-# Copyright 2012-2017 The Meson development team
+# Copyright 2012-2018 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,25 +12,31 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from . import mlog
 import pickle, os, uuid
+import sys
 from pathlib import PurePath
 from collections import OrderedDict
-from .mesonlib import MesonException, commonpath
+from .mesonlib import MesonException
 from .mesonlib import default_libdir, default_libexecdir, default_prefix
 import ast
 
-version = '0.43.0'
+version = '0.45.1'
 backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
 
+default_yielding = False
+
 class UserOption:
-    def __init__(self, name, description, choices):
+    def __init__(self, name, description, choices, yielding):
         super().__init__()
         self.name = name
         self.choices = choices
         self.description = description
-
-    def parse_string(self, valuestring):
-        return valuestring
+        if yielding is None:
+            yielding = default_yielding
+        if not isinstance(yielding, bool):
+            raise MesonException('Value of "yielding" must be a boolean.')
+        self.yielding = yielding
 
     # Check that the input is a valid value and return the
     # "cleaned" or "native" version. For example the Boolean
@@ -39,8 +45,8 @@ class UserOption:
         raise RuntimeError('Derived option class did not override validate_value.')
 
 class UserStringOption(UserOption):
-    def __init__(self, name, description, value, choices=None):
-        super().__init__(name, description, choices)
+    def __init__(self, name, description, value, choices=None, yielding=None):
+        super().__init__(name, description, choices, yielding)
         self.set_value(value)
 
     def validate(self, value):
@@ -56,8 +62,8 @@ class UserStringOption(UserOption):
         return value
 
 class UserBooleanOption(UserOption):
-    def __init__(self, name, description, value):
-        super().__init__(name, description, [True, False])
+    def __init__(self, name, description, value, yielding=None):
+        super().__init__(name, description, [True, False], yielding)
         self.set_value(value)
 
     def tobool(self, thing):
@@ -72,13 +78,6 @@ class UserBooleanOption(UserOption):
     def set_value(self, newvalue):
         self.value = self.tobool(newvalue)
 
-    def parse_string(self, valuestring):
-        if valuestring == 'false':
-            return False
-        if valuestring == 'true':
-            return True
-        raise MesonException('Value "%s" for boolean option "%s" is not a boolean.' % (valuestring, self.name))
-
     def __bool__(self):
         return self.value
 
@@ -86,11 +85,17 @@ class UserBooleanOption(UserOption):
         return self.tobool(value)
 
 class UserIntegerOption(UserOption):
-    def __init__(self, name, description, min_value, max_value, value):
-        super().__init__(name, description, [True, False])
+    def __init__(self, name, description, min_value, max_value, value, yielding=None):
+        super().__init__(name, description, [True, False], yielding)
         self.min_value = min_value
         self.max_value = max_value
         self.set_value(value)
+        c = []
+        if min_value is not None:
+            c.append('>=' + str(min_value))
+        if max_value is not None:
+            c.append('<=' + str(max_value))
+        self.choices = ', '.join(c)
 
     def set_value(self, newvalue):
         if isinstance(newvalue, str):
@@ -109,15 +114,12 @@ class UserIntegerOption(UserOption):
         except:
             raise MesonException('Value string "%s" is not convertable to an integer.' % valuestring)
 
-    def parse_string(self, valuestring):
-        return self.toint(valuestring)
-
     def validate_value(self, value):
         return self.toint(value)
 
 class UserComboOption(UserOption):
-    def __init__(self, name, description, choices, value):
-        super().__init__(name, description, choices)
+    def __init__(self, name, description, choices, value, yielding=None):
+        super().__init__(name, description, choices, yielding)
         if not isinstance(self.choices, list):
             raise MesonException('Combo choices must be an array.')
         for i in self.choices:
@@ -136,27 +138,46 @@ class UserComboOption(UserOption):
             raise MesonException('Value %s not one of accepted values.' % value)
         return value
 
-class UserStringArrayOption(UserOption):
+class UserArrayOption(UserOption):
     def __init__(self, name, description, value, **kwargs):
-        super().__init__(name, description, kwargs.get('choices', []))
-        self.set_value(value)
+        super().__init__(name, description, kwargs.get('choices', []), yielding=kwargs.get('yielding', None))
+        self.set_value(value, user_input=False)
 
-    def validate(self, value):
-        if isinstance(value, str):
-            if not value.startswith('['):
-                raise MesonException('Valuestring does not define an array: ' + value)
-            newvalue = ast.literal_eval(value)
+    def validate(self, value, user_input):
+        # User input is for options defined on the command line (via -D
+        # options). Users can put their input in as a comma separated
+        # string, but for defining options in meson_options.txt the format
+        # should match that of a combo
+        if not user_input:
+            if isinstance(value, str):
+                if not value.startswith('['):
+                    raise MesonException('Valuestring does not define an array: ' + value)
+                newvalue = ast.literal_eval(value)
+            else:
+                newvalue = value
         else:
-            newvalue = value
+            assert isinstance(value, str)
+            if value.startswith('['):
+                newvalue = ast.literal_eval(value)
+            else:
+                newvalue = [v.strip() for v in value.split(',')]
+                if len(set(newvalue)) != len(newvalue):
+                    mlog.log(mlog.red('DEPRECATION:'), '''Duplicated values in an array type is deprecated.
+This will become a hard error in the future.''')
         if not isinstance(newvalue, list):
             raise MesonException('"{0}" should be a string array, but it is not'.format(str(newvalue)))
         for i in newvalue:
             if not isinstance(i, str):
                 raise MesonException('String array element "{0}" is not a string.'.format(str(newvalue)))
+        if self.choices:
+            bad = [x for x in newvalue if x not in self.choices]
+            if bad:
+                raise MesonException('Options "{}" are not in allowed choices: "{}"'.format(
+                    ', '.join(bad), ', '.join(self.choices)))
         return newvalue
 
-    def set_value(self, newvalue):
-        self.value = self.validate(newvalue)
+    def set_value(self, newvalue, user_input=True):
+        self.value = self.validate(newvalue, user_input)
 
     def validate_value(self, value):
         self.validate(value)
@@ -184,10 +205,7 @@ class CoreData:
         self.external_preprocess_args = {} # CPPFLAGS only
         self.external_args = {} # CPPFLAGS + CFLAGS
         self.external_link_args = {} # CFLAGS + LDFLAGS (with MSVC: only LDFLAGS)
-        if options.cross_file is not None:
-            self.cross_file = os.path.join(os.getcwd(), options.cross_file)
-        else:
-            self.cross_file = None
+        self.cross_file = self.__load_cross_file(options.cross_file)
         self.wrap_mode = options.wrap_mode
         self.compilers = OrderedDict()
         self.cross_compilers = OrderedDict()
@@ -196,6 +214,46 @@ class CoreData:
         # Only to print a warning if it changes between Meson invocations.
         self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
 
+    @staticmethod
+    def __load_cross_file(filename):
+        """Try to load the cross file.
+
+        If the filename is None return None. If the filename is an absolute
+        (after resolving variables and ~), return that absolute path. Next,
+        check if the file is relative to the current source dir. If the path
+        still isn't resolved do the following:
+            Windows:
+                - Error
+            *:
+                - $XDG_DATA_HOME/meson/cross (or ~/.local/share/meson/cross if
+                  undefined)
+                - $XDG_DATA_DIRS/meson/cross (or
+                  /usr/local/share/meson/cross:/usr/share/meson/cross if undefined)
+                - Error
+
+        Non-Windows follows the Linux path and will honor XDG_* if set. This
+        simplifies the implementation somewhat.
+        """
+        if filename is None:
+            return None
+        filename = os.path.expanduser(os.path.expandvars(filename))
+        if os.path.isabs(filename):
+            return filename
+        path_to_try = os.path.abspath(filename)
+        if os.path.exists(path_to_try):
+            return path_to_try
+        if sys.platform != 'win32':
+            paths = [
+                os.environ.get('XDG_DATA_HOME', os.path.expanduser('~/.local/share')),
+            ] + os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
+            for path in paths:
+                path_to_try = os.path.join(path, 'meson', 'cross', filename)
+                if os.path.exists(path_to_try):
+                    return path_to_try
+            raise MesonException('Cannot find specified cross file: ' + filename)
+
+        raise MesonException('Cannot find specified cross file: ' + filename)
+
     def sanitize_prefix(self, prefix):
         if not os.path.isabs(prefix):
             raise MesonException('prefix value {!r} must be an absolute path'
@@ -205,6 +263,10 @@ class CoreData:
             # string is of type 'C:\' because 'C:' is not an absolute path.
             if len(prefix) == 3 and prefix[1] == ':':
                 pass
+            # If prefix is a single character, preserve it since it is
+            # the root directory.
+            elif len(prefix) == 1:
+                pass
             else:
                 prefix = prefix[:-1]
         return prefix
@@ -224,7 +286,7 @@ class CoreData:
             # commonpath will always return a path in the native format, so we
             # must use pathlib.PurePath to do the same conversion before
             # comparing.
-            if commonpath([value, prefix]) != str(PurePath(prefix)):
+            if os.path.commonpath([value, prefix]) != str(PurePath(prefix)):
                 m = 'The value of the {!r} option is {!r} which must be a ' \
                     'subdir of the prefix {!r}.\nNote that if you pass a ' \
                     'relative path, it is assumed to be a subdir of prefix.'
@@ -245,7 +307,7 @@ class CoreData:
                 value = self.sanitize_dir_option_value(options.prefix, key, value)
                 setattr(options, key, value)
             else:
-                value = get_builtin_option_default(key)
+                value = get_builtin_option_default(key, prefix=options.prefix)
             args = [key] + builtin_options[key][1:-1] + [value]
             self.builtins[key] = builtin_options[key][0](*args)
 
@@ -278,7 +340,8 @@ class CoreData:
                 return opt.validate_value(override_value)
         raise MesonException('Tried to validate unknown option %s.' % option_name)
 
-def load(filename):
+def load(build_dir):
+    filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
     try:
         with open(filename, 'rb') as f:
@@ -292,7 +355,8 @@ def load(filename):
                              (obj.version, version))
     return obj
 
-def save(obj, filename):
+def save(obj, build_dir):
+    filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     if obj.version != version:
         raise MesonException('Fatal version mismatch corruption.')
     with open(filename, 'wb') as f:
@@ -321,11 +385,19 @@ def get_builtin_option_description(optname):
     else:
         raise RuntimeError('Tried to get the description for an unknown builtin option \'%s\'.' % optname)
 
-def get_builtin_option_default(optname):
+def get_builtin_option_default(optname, prefix='', noneIfSuppress=False):
     if is_builtin_option(optname):
         o = builtin_options[optname]
         if o[0] == UserComboOption:
             return o[3]
+        if optname in builtin_dir_noprefix_options:
+            if noneIfSuppress:
+                # Return None if argparse defaulting should be suppressed for
+                # this option (so we can determine the default later based on
+                # prefix)
+                return None
+            elif prefix in builtin_dir_noprefix_options[optname]:
+                return builtin_dir_noprefix_options[optname][prefix]
         return o[2]
     else:
         raise RuntimeError('Tried to get the default value for an unknown builtin option \'%s\'.' % optname)
@@ -344,15 +416,6 @@ builtin_options = {
     'mandir':     [UserStringOption, 'Manual page directory.', 'share/man'],
     'infodir':    [UserStringOption, 'Info page directory.', 'share/info'],
     'localedir':  [UserStringOption, 'Locale data directory.', 'share/locale'],
-    # sysconfdir, localstatedir and sharedstatedir are a bit special. These defaults to ${prefix}/etc,
-    # ${prefix}/var and ${prefix}/com but nobody uses that. Instead they always set it
-    # manually to /etc, /var and /var/lib. This default values is thus pointless and not really used
-    # but we set it to this for consistency with other systems.
-    #
-    # Projects installing to sysconfdir, localstatedir or sharedstatedir probably want
-    # to set the following in project():
-    #
-    # default_options : ['sysconfdir=/etc', 'localstatedir=/var', 'sharedstatedir=/var/lib']
     'sysconfdir':      [UserStringOption, 'Sysconf data directory.', 'etc'],
     'localstatedir':   [UserStringOption, 'Localstate data directory.', 'var'],
     'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory.', 'com'],
@@ -365,8 +428,13 @@ builtin_options = {
     'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests.", True],
 }
 
-# Installation directories that can reside in a path outside of the prefix
-builtin_dir_noprefix_options = {'sysconfdir', 'localstatedir', 'sharedstatedir'}
+# Special prefix-dependent defaults for installation directories that reside in
+# a path outside of the prefix in FHS and common usage.
+builtin_dir_noprefix_options = {
+    'sysconfdir':     {'/usr': '/etc'},
+    'localstatedir':  {'/usr': '/var',     '/usr/local': '/var/local'},
+    'sharedstatedir': {'/usr': '/var/lib', '/usr/local': '/var/local/lib'},
+}
 
 forbidden_target_names = {'clean': None,
                           'clean-ctlist': None,
