@@ -18,7 +18,7 @@
 from . import mparser, mesonlib, mlog
 from . import environment, dependencies
 
-import os, copy, re
+import os, copy, re, types
 from functools import wraps
 
 # Decorators for method calls.
@@ -63,17 +63,19 @@ class permittedKwargs:
     def __call__(self, f):
         @wraps(f)
         def wrapped(s, node_or_state, args, kwargs):
+            loc = types.SimpleNamespace()
             if hasattr(s, 'subdir'):
-                subdir = s.subdir
-                lineno = s.current_lineno
+                loc.subdir = s.subdir
+                loc.lineno = s.current_lineno
             elif hasattr(node_or_state, 'subdir'):
-                subdir = node_or_state.subdir
-                lineno = node_or_state.current_lineno
+                loc.subdir = node_or_state.subdir
+                loc.lineno = node_or_state.current_lineno
+            else:
+                loc = None
             for k in kwargs:
                 if k not in self.permitted:
-                    fname = os.path.join(subdir, environment.build_filename)
-                    mlog.warning('''Passed invalid keyword argument "%s" in %s line %d.
-This will become a hard error in the future.''' % (k, fname, lineno))
+                    mlog.warning('''Passed invalid keyword argument "{}".'''.format(k), location=loc)
+                    mlog.warning('This will become a hard error in the future.')
             return f(s, node_or_state, args, kwargs)
         return wrapped
 
@@ -263,6 +265,11 @@ class InterpreterBase:
         if not isinstance(node.elseblock, mparser.EmptyNode):
             self.evaluate_codeblock(node.elseblock)
 
+    def validate_comparison_types(self, val1, val2):
+        if type(val1) != type(val2):
+            return False
+        return True
+
     def evaluate_comparison(self, node):
         val1 = self.evaluate_statement(node.left)
         if is_disabler(val1):
@@ -270,11 +277,19 @@ class InterpreterBase:
         val2 = self.evaluate_statement(node.right)
         if is_disabler(val2):
             return val2
+        valid = self.validate_comparison_types(val1, val2)
+        # Ordering comparisons of different types isn't allowed since PR #1810
+        # (0.41.0).  Since PR #2884 we also warn about equality comparisons of
+        # different types, which will one day become an error.
+        if not valid and (node.ctype == '==' or node.ctype == '!='):
+            mlog.warning('''Trying to compare values of different types ({}, {}) using {}.
+The result of this is undefined and will become a hard error in a future Meson release.'''
+                         .format(type(val1).__name__, type(val2).__name__, node.ctype), location=node)
         if node.ctype == '==':
             return val1 == val2
         elif node.ctype == '!=':
             return val1 != val2
-        elif not isinstance(val1, type(val2)):
+        elif not valid:
             raise InterpreterException(
                 'Values of different types ({}, {}) cannot be compared using {}.'.format(type(val1).__name__,
                                                                                          type(val2).__name__,
