@@ -15,13 +15,11 @@
 '''This module provides helper functions for Gnome/GLib related
 functionality such as gobject-introspection, gresources and gtk-doc'''
 
-from .. import build
 import os
 import copy
 import subprocess
-from . import ModuleReturnValue
-from ..mesonlib import MesonException, OrderedSet, Popen_safe, extract_as_list
-from ..dependencies import Dependency, PkgConfigDependency, InternalDependency
+
+from .. import build
 from .. import mlog
 from .. import mesonlib
 from .. import compilers
@@ -29,6 +27,9 @@ from .. import interpreter
 from . import GResourceTarget, GResourceHeaderTarget, GirTarget, TypelibTarget, VapiTarget
 from . import find_program, get_include_args
 from . import ExtensionModule
+from . import ModuleReturnValue
+from ..mesonlib import MesonException, OrderedSet, Popen_safe, extract_as_list
+from ..dependencies import Dependency, PkgConfigDependency, InternalDependency
 from ..interpreterbase import noKwargs, permittedKwargs
 
 # gresource compilation is broken due to the way
@@ -192,7 +193,7 @@ class GnomeModule(ExtensionModule):
             depfile = kwargs['output'] + '.d'
             kwargs['depfile'] = depfile
             kwargs['command'] = copy.copy(cmd) + ['--dependency-file', '@DEPFILE@']
-        target_c = GResourceTarget(name, state.subdir, kwargs)
+        target_c = GResourceTarget(name, state.subdir, state.subproject, kwargs)
 
         if gresource: # Only one target for .gresource files
             return ModuleReturnValue(target_c, [target_c])
@@ -210,7 +211,7 @@ class GnomeModule(ExtensionModule):
             h_kwargs['install'] = install_header
             h_kwargs['install_dir'] = kwargs.get('install_dir',
                                                  state.environment.coredata.get_builtin_option('includedir'))
-        target_h = GResourceHeaderTarget(args[0] + '_h', state.subdir, h_kwargs)
+        target_h = GResourceHeaderTarget(args[0] + '_h', state.subdir, state.subproject, h_kwargs)
         rv = [target_c, target_h]
         return ModuleReturnValue(rv, rv)
 
@@ -220,9 +221,10 @@ class GnomeModule(ExtensionModule):
                input_file,
                '--generate-dependencies']
 
+        # Prefer generated files over source files
+        cmd += ['--sourcedir', state.subdir] # Current build dir
         for source_dir in source_dirs:
             cmd += ['--sourcedir', os.path.join(state.subdir, source_dir)]
-        cmd += ['--sourcedir', state.subdir] # Current dir
 
         pc, stdout, stderr = Popen_safe(cmd, cwd=state.environment.get_source_dir())
         if pc.returncode != 0:
@@ -232,33 +234,17 @@ class GnomeModule(ExtensionModule):
 
         dep_files = stdout.split('\n')[:-1]
 
-        # In generate-dependencies mode, glib-compile-resources doesn't raise
-        # an error for missing resources but instead prints whatever filename
-        # was listed in the input file.  That's good because it means we can
-        # handle resource files that get generated as part of the build, as
-        # follows.
-        #
-        # If there are multiple generated resource files with the same basename
-        # then this code will get confused.
-
-        def exists_in_srcdir(f):
-            return os.path.exists(os.path.join(state.environment.get_source_dir(), f))
-        missing_dep_files = [f for f in dep_files if not exists_in_srcdir(f)]
-
         depends = []
         subdirs = []
-        for missing in missing_dep_files:
-            found = False
-            missing_basename = os.path.basename(missing)
-
+        for resfile in dep_files[:]:
+            resbasename = os.path.basename(resfile)
             for dep in dependencies:
                 if hasattr(dep, 'held_object'):
                     dep = dep.held_object
                 if isinstance(dep, mesonlib.File):
-                    if dep.fname != missing_basename:
+                    if dep.fname != resbasename:
                         continue
-                    found = True
-                    dep_files.remove(missing)
+                    dep_files.remove(resfile)
                     dep_files.append(dep)
                     subdirs.append(dep.subdir)
                     break
@@ -266,30 +252,34 @@ class GnomeModule(ExtensionModule):
                     fname = None
                     outputs = {(o, os.path.basename(o)) for o in dep.get_outputs()}
                     for o, baseo in outputs:
-                        if baseo == missing_basename:
+                        if baseo == resbasename:
                             fname = o
                             break
                     if fname is not None:
-                        found = True
-                        dep_files.remove(missing)
-                        dep_files.append(
-                            mesonlib.File(
-                                is_built=True,
-                                subdir=dep.get_subdir(),
-                                fname=fname))
+                        dep_files.remove(resfile)
                         depends.append(dep)
                         subdirs.append(dep.get_subdir())
                         break
-                else:
-                    raise RuntimeError('Unreachable code.')
-
-            if not found:
-                raise MesonException(
-                    'Resource "%s" listed in "%s" was not found. If this is a '
-                    'generated file, pass the target that generates it to '
-                    'gnome.compile_resources() using the "dependencies" '
-                    'keyword argument.' % (missing, input_file))
-
+            else:
+                # In generate-dependencies mode, glib-compile-resources doesn't raise
+                # an error for missing resources but instead prints whatever filename
+                # was listed in the input file.  That's good because it means we can
+                # handle resource files that get generated as part of the build, as
+                # follows.
+                #
+                # If there are multiple generated resource files with the same basename
+                # then this code will get confused.
+                try:
+                    f = mesonlib.File.from_source_file(state.environment.get_source_dir(),
+                                                       ".", resfile)
+                except MesonException:
+                    raise MesonException(
+                        'Resource "%s" listed in "%s" was not found. If this is a '
+                        'generated file, pass the target that generates it to '
+                        'gnome.compile_resources() using the "dependencies" '
+                        'keyword argument.' % (resfile, input_file))
+                dep_files.remove(resfile)
+                dep_files.append(f)
         return dep_files, depends, subdirs
 
     def _get_link_args(self, state, lib, depends=None, include_rpath=False,
@@ -370,7 +360,7 @@ class GnomeModule(ExtensionModule):
                     ldflags.update([lib])
 
                 if isinstance(dep, PkgConfigDependency):
-                    girdir = dep.get_pkgconfig_variable("girdir")
+                    girdir = dep.get_pkgconfig_variable("girdir", {'default': ''})
                     if girdir:
                         gi_includes.update([girdir])
             elif isinstance(dep, (build.StaticLibrary, build.SharedLibrary)):
@@ -380,7 +370,7 @@ class GnomeModule(ExtensionModule):
                 continue
 
         if gir_has_extra_lib_arg() and use_gir_args:
-            fixed_ldflags = set()
+            fixed_ldflags = OrderedSet()
             for ldflag in ldflags:
                 if ldflag.startswith("-l"):
                     fixed_ldflags.add(ldflag.replace('-l', '--extra-library=', 1))
@@ -561,7 +551,7 @@ class GnomeModule(ExtensionModule):
                         if subdir not in typelib_includes:
                             typelib_includes.append(subdir)
             elif isinstance(dep, PkgConfigDependency):
-                girdir = dep.get_pkgconfig_variable("girdir")
+                girdir = dep.get_pkgconfig_variable("girdir", {'default': ''})
                 if girdir and girdir not in typelib_includes:
                     typelib_includes.append(girdir)
         # ldflags will be misinterpreted by gir scanner (showing
@@ -573,6 +563,7 @@ class GnomeModule(ExtensionModule):
         ldflags += list(dep_ldflags)
         scan_command += ['--cflags-begin']
         scan_command += cflags
+        scan_command += state.environment.coredata.external_args[lang]
         scan_command += ['--cflags-end']
         # need to put our output directory first as we need to use the
         # generated libraries instead of any possibly installed system/prefix
@@ -603,6 +594,11 @@ class GnomeModule(ExtensionModule):
                 d = os.path.join(state.environment.get_build_dir(), d)
                 scan_command.append('-L' + d)
             scan_command += ['--library', libname]
+
+        for link_arg in state.environment.coredata.external_link_args[lang]:
+            if link_arg.startswith('-L'):
+                scan_command.append(link_arg)
+
         scankwargs = {'output': girfile,
                       'command': scan_command,
                       'depends': depends}
@@ -612,7 +608,7 @@ class GnomeModule(ExtensionModule):
                                                    os.path.join(state.environment.get_datadir(), 'gir-1.0'))
         if 'build_by_default' in kwargs:
             scankwargs['build_by_default'] = kwargs['build_by_default']
-        scan_target = GirTarget(girfile, state.subdir, scankwargs)
+        scan_target = GirTarget(girfile, state.subdir, state.subproject, scankwargs)
 
         typelib_output = '%s-%s.typelib' % (ns, nsversion)
         typelib_cmd = [gicompiler, scan_target, '--output', '@OUTPUT@']
@@ -630,11 +626,11 @@ class GnomeModule(ExtensionModule):
                                                        os.path.join(state.environment.get_libdir(), 'girepository-1.0'))
         if 'build_by_default' in kwargs:
             typelib_kwargs['build_by_default'] = kwargs['build_by_default']
-        typelib_target = TypelibTarget(typelib_output, state.subdir, typelib_kwargs)
+        typelib_target = TypelibTarget(typelib_output, state.subdir, state.subproject, typelib_kwargs)
         rv = [scan_target, typelib_target]
         return ModuleReturnValue(rv, rv)
 
-    @permittedKwargs({'build_by_default'})
+    @permittedKwargs({'build_by_default', 'depend_files'})
     def compile_schemas(self, state, args, kwargs):
         if args:
             raise MesonException('Compile_schemas does not take positional arguments.')
@@ -649,8 +645,8 @@ class GnomeModule(ExtensionModule):
         if state.subdir == '':
             targetname = 'gsettings-compile'
         else:
-            targetname = 'gsettings-compile-' + state.subdir
-        target_g = build.CustomTarget(targetname, state.subdir, kwargs)
+            targetname = 'gsettings-compile-' + state.subdir.replace('/', '_')
+        target_g = build.CustomTarget(targetname, state.subdir, state.subproject, kwargs)
         return ModuleReturnValue(target_g, [target_g])
 
     @permittedKwargs({'sources', 'media', 'symlink_media', 'languages'})
@@ -705,7 +701,7 @@ This will become a hard error in the future.''')
             '--sources=' + source_str,
         ]
         pottarget = build.RunTarget('help-' + project_id + '-pot', potargs[0],
-                                    potargs[1:], [], state.subdir)
+                                    potargs[1:], [], state.subdir, state.subproject)
 
         poargs = state.environment.get_build_command() + [
             '--internal', 'yelphelper', 'update-po',
@@ -715,7 +711,7 @@ This will become a hard error in the future.''')
             '--langs=' + '@@'.join(langs),
         ]
         potarget = build.RunTarget('help-' + project_id + '-update-po', poargs[0],
-                                   poargs[1:], [], state.subdir)
+                                   poargs[1:], [], state.subdir, state.subproject)
 
         rv = [inscript, pottarget, potarget]
         return ModuleReturnValue(None, rv)
@@ -723,7 +719,8 @@ This will become a hard error in the future.''')
     @permittedKwargs({'main_xml', 'main_sgml', 'src_dir', 'dependencies', 'install',
                       'install_dir', 'scan_args', 'scanobjs_args', 'gobject_typesfile',
                       'fixxref_args', 'html_args', 'html_assets', 'content_files',
-                      'mkdb_args', 'ignore_headers'})
+                      'mkdb_args', 'ignore_headers', 'include_directories',
+                      'namespace', 'mode', 'expand_content_files'})
     def gtkdoc(self, state, args, kwargs):
         if len(args) != 1:
             raise MesonException('Gtkdoc must have one positional argument.')
@@ -783,12 +780,39 @@ This will become a hard error in the future.''')
         args += self._unpack_args('--fixxrefargs=', 'fixxref_args', kwargs)
         args += self._unpack_args('--mkdbargs=', 'mkdb_args', kwargs)
         args += self._unpack_args('--html-assets=', 'html_assets', kwargs, state)
-        args += self._unpack_args('--content-files=', 'content_files', kwargs, state)
+
+        depends = []
+        content_files = []
+        for s in mesonlib.extract_as_list(kwargs, 'content_files'):
+            if hasattr(s, 'held_object'):
+                s = s.held_object
+            if isinstance(s, (build.CustomTarget, build.CustomTargetIndex)):
+                depends.append(s)
+                content_files.append(os.path.join(state.environment.get_build_dir(),
+                                                  state.backend.get_target_dir(s),
+                                                  s.get_outputs()[0]))
+            elif isinstance(s, mesonlib.File):
+                content_files.append(s.rel_to_builddir(state.build_to_src))
+            elif isinstance(s, build.GeneratedList):
+                depends.append(s)
+                for gen_src in s.get_outputs():
+                    content_files.append(os.path.join(state.environment.get_source_dir(),
+                                                      state.subdir,
+                                                      gen_src))
+            elif isinstance(s, str):
+                content_files.append(os.path.join(state.environment.get_source_dir(),
+                                                  state.subdir,
+                                                  s))
+            else:
+                raise MesonException(
+                    'Invalid object type: {!r}'.format(s.__class__.__name__))
+        args += ['--content-files=' + '@@'.join(content_files)]
+
         args += self._unpack_args('--expand-content-files=', 'expand_content_files', kwargs, state)
         args += self._unpack_args('--ignore-headers=', 'ignore_headers', kwargs)
         args += self._unpack_args('--installdir=', 'install_dir', kwargs, state)
         args += self._get_build_args(kwargs, state)
-        res = [build.RunTarget(targetname, command[0], command[1:] + args, [], state.subdir)]
+        res = [build.RunTarget(targetname, command[0], command[1:] + args, depends, state.subdir, state.subproject)]
         if kwargs.get('install', True):
             res.append(build.RunScript(command, args))
         return ModuleReturnValue(None, res)
@@ -803,6 +827,8 @@ This will become a hard error in the future.''')
                 raise MesonException(
                     'Gir include dirs should be include_directories().')
         cflags.update(get_include_args(inc_dirs))
+        cflags.update(state.environment.coredata.external_args['c'])
+        ldflags.update(state.environment.coredata.external_link_args['c'])
         if cflags:
             args += ['--cflags=%s' % ' '.join(cflags)]
         if ldflags:
@@ -832,7 +858,7 @@ This will become a hard error in the future.''')
         args = []
         for i in new_args:
             if expend_file_state and isinstance(i, mesonlib.File):
-                i = os.path.join(expend_file_state.environment.get_build_dir(), i.subdir, i.fname)
+                i = i.absolute_path(expend_file_state.environment.get_source_dir(), expend_file_state.environment.get_build_dir())
             elif not isinstance(i, str):
                 raise MesonException(kwarg_name + ' values must be strings.')
             args.append(i)
@@ -885,7 +911,7 @@ This will become a hard error in the future.''')
                          }
         if 'build_by_default' in kwargs:
             custom_kwargs['build_by_default'] = kwargs['build_by_default']
-        ct = build.CustomTarget(target_name, state.subdir, custom_kwargs)
+        ct = build.CustomTarget(target_name, state.subdir, state.subproject, custom_kwargs)
         return ModuleReturnValue(ct, [ct])
 
     @permittedKwargs({'sources', 'c_template', 'h_template', 'install_header', 'install_dir',
@@ -1101,7 +1127,7 @@ G_END_DECLS'''
             'command': cmd
         }
         custom_kwargs.update(kwargs)
-        return build.CustomTarget(output, state.subdir, custom_kwargs,
+        return build.CustomTarget(output, state.subdir, state.subproject, custom_kwargs,
                                   # https://github.com/mesonbuild/meson/issues/973
                                   absolute_paths=True)
 
@@ -1171,7 +1197,7 @@ G_END_DECLS'''
             # Silence any warnings about missing prototypes
             custom_kwargs['command'] += ['--include-header', header_file]
         custom_kwargs['output'] = output + '.c'
-        body = build.CustomTarget(output + '_c', state.subdir, custom_kwargs)
+        body = build.CustomTarget(output + '_c', state.subdir, state.subproject, custom_kwargs)
 
         custom_kwargs['install'] = install_header
         if install_dir is not None:
@@ -1180,7 +1206,7 @@ G_END_DECLS'''
             cmd += ['--pragma-once']
         custom_kwargs['command'] = cmd + ['--header', '@INPUT@']
         custom_kwargs['output'] = header_file
-        header = build.CustomTarget(output + '_h', state.subdir, custom_kwargs)
+        header = build.CustomTarget(output + '_h', state.subdir, state.subproject, custom_kwargs)
 
         rv = [body, header]
         return ModuleReturnValue(rv, rv)
@@ -1316,7 +1342,7 @@ G_END_DECLS'''
             # We shouldn't need this locally but we install it
             deps_target = self._generate_deps(state, library, vapi_packages, install_dir)
             created_values.append(deps_target)
-        vapi_target = VapiTarget(vapi_output, state.subdir, custom_kwargs)
+        vapi_target = VapiTarget(vapi_output, state.subdir, state.subproject, custom_kwargs)
 
         # So to try our best to get this to just work we need:
         # - link with with the correct library
