@@ -31,7 +31,7 @@ import argparse
 import xml.etree.ElementTree as ET
 import time
 import multiprocessing
-import concurrent.futures as conc
+from concurrent.futures import ProcessPoolExecutor
 import re
 from run_unittests import get_fake_options, run_configure
 
@@ -59,50 +59,6 @@ class TestResult:
         self.buildtime = buildtime
         self.testtime = testtime
 
-class DummyFuture(conc.Future):
-    '''
-    Dummy Future implementation that executes the provided function when you
-    ask for the result. Used on platforms where sem_open() is not available:
-    MSYS2, OpenBSD, etc: https://bugs.python.org/issue3770
-    '''
-    def set_function(self, fn, *args, **kwargs):
-        self.fn = fn
-        self.fn_args = args
-        self.fn_kwargs = kwargs
-
-    def result(self, **kwargs):
-        try:
-            result = self.fn(*self.fn_args, **self.fn_kwargs)
-        except BaseException as e:
-            self.set_exception(e)
-        else:
-            self.set_result(result)
-        return super().result(**kwargs)
-
-
-class DummyExecutor(conc.Executor):
-    '''
-    Dummy single-thread 'concurrent' executor for use on platforms where
-    sem_open is not available: https://bugs.python.org/issue3770
-    '''
-
-    def __init__(self):
-        from threading import Lock
-        self._shutdown = False
-        self._shutdownLock = Lock()
-
-    def submit(self, fn, *args, **kwargs):
-        with self._shutdownLock:
-            if self._shutdown:
-                raise RuntimeError('Cannot schedule new futures after shutdown')
-            f = DummyFuture()
-            f.set_function(fn, *args, **kwargs)
-            return f
-
-    def shutdown(self, wait=True):
-        with self._shutdownLock:
-            self._shutdown = True
-
 
 class AutoDeletedDir:
     def __init__(self, d):
@@ -121,7 +77,8 @@ class AutoDeletedDir:
 
 failing_logs = []
 print_debug = 'MESON_PRINT_TEST_OUTPUT' in os.environ
-do_debug = not {'MESON_PRINT_TEST_OUTPUT', 'TRAVIS', 'APPVEYOR'}.isdisjoint(os.environ)
+under_ci = not {'TRAVIS', 'APPVEYOR'}.isdisjoint(os.environ)
+do_debug = under_ci or print_debug
 no_meson_log_msg = 'No meson-log.txt found.'
 
 system_compiler = None
@@ -185,6 +142,7 @@ def get_relative_files_list_from_dir(fromdir):
 def platform_fix_name(fname, compiler):
     if '?lib' in fname:
         if mesonlib.is_cygwin():
+            fname = re.sub(r'lib/\?lib(.*)\.so$', r'bin/cyg\1.dll', fname)
             fname = re.sub(r'\?lib(.*)\.dll$', r'cyg\1.dll', fname)
         else:
             fname = re.sub(r'\?lib', 'lib', fname)
@@ -312,10 +270,19 @@ def parse_test_args(testdir):
         pass
     return args
 
+# Build directory name must be the same so CCache works over
+# consecutive invocations.
+def create_deterministic_builddir(src_dir):
+    import hashlib
+    rel_dirname = 'b ' + hashlib.sha256(src_dir.encode(errors='ignore')).hexdigest()[0:10]
+    os.mkdir(rel_dirname)
+    abs_pathname = os.path.join(os.getcwd(), rel_dirname)
+    return abs_pathname
+
 def run_test(skipped, testdir, extra_args, compiler, backend, flags, commands, should_fail):
     if skipped:
         return None
-    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
+    with AutoDeletedDir(create_deterministic_builddir(testdir)) as build_dir:
         with AutoDeletedDir(tempfile.mkdtemp(prefix='i ', dir=os.getcwd())) as install_dir:
             try:
                 return _run_test(testdir, build_dir, install_dir, extra_args, compiler, backend, flags, commands, should_fail)
@@ -355,9 +322,12 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
         mesonlog = no_meson_log_msg
     gen_time = time.time() - gen_start
     if should_fail == 'meson':
-        if returncode != 0:
+        if returncode == 1:
             return TestResult('', BuildStep.configure, stdo, stde, mesonlog, gen_time)
-        return TestResult('Test that should have failed succeeded', BuildStep.configure, stdo, stde, mesonlog, gen_time)
+        elif returncode != 0:
+            return TestResult('Test exited with unexpected status {}'.format(returncode), BuildStep.configure, stdo, stde, mesonlog, gen_time)
+        else:
+            return TestResult('Test that should have failed succeeded', BuildStep.configure, stdo, stde, mesonlog, gen_time)
     if returncode != 0:
         return TestResult('Generating the build system failed.', BuildStep.configure, stdo, stde, mesonlog, gen_time)
     # Touch the meson.build file to force a regenerate so we can test that
@@ -438,20 +408,28 @@ def have_objc_compiler():
         env = environment.Environment(None, build_dir, None, get_fake_options('/'), [])
         try:
             objc_comp = env.detect_objc_compiler(False)
-        except:
+        except mesonlib.MesonException:
             return False
         if not objc_comp:
             return False
         try:
             objc_comp.sanity_check(env.get_scratch_dir(), env)
-            objcpp_comp = env.detect_objc_compiler(False)
-        except:
+        except mesonlib.MesonException:
+            return False
+    return True
+
+def have_objcpp_compiler():
+    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
+        env = environment.Environment(None, build_dir, None, get_fake_options('/'), [])
+        try:
+            objcpp_comp = env.detect_objcpp_compiler(False)
+        except mesonlib.MesonException:
             return False
         if not objcpp_comp:
             return False
         try:
             objcpp_comp.sanity_check(env.get_scratch_dir(), env)
-        except:
+        except mesonlib.MesonException:
             return False
     return True
 
@@ -460,6 +438,55 @@ def have_java():
         return True
     return False
 
+def skippable(suite, test):
+    if not under_ci:
+        return True
+
+    if not suite.endswith('frameworks'):
+        return True
+
+    # gtk-doc test may be skipped, pending upstream fixes for spaces in
+    # filenames landing in the distro used for CI
+    if test.endswith('10 gtk-doc'):
+        return True
+
+    # No frameworks test should be skipped on linux CI, as we expect all
+    # prerequisites to be installed
+    if mesonlib.is_linux():
+        return False
+
+    # Boost test should only be skipped for windows CI build matrix entries
+    # which don't define BOOST_ROOT
+    if test.endswith('1 boost'):
+        if mesonlib.is_windows():
+            return 'BOOST_ROOT' not in os.environ
+        return False
+
+    # Other framework tests are allowed to be skipped on other platforms
+    return True
+
+def skip_csharp(backend):
+    if backend is not Backend.ninja:
+        return True
+    if not shutil.which('resgen'):
+        return True
+    if shutil.which('mcs'):
+        return False
+    if shutil.which('csc'):
+        # Only support VS2017 for now. Earlier versions fail
+        # under CI in mysterious ways.
+        try:
+            stdo = subprocess.check_output(['csc', '/version'])
+        except subprocess.CalledProcessError:
+            return True
+        # Having incrementing version numbers would be too easy.
+        # Microsoft reset the versioning back to 1.0 (from 4.x)
+        # when they got the Roslyn based compiler. Thus there
+        # is NO WAY to reliably do version number comparisons.
+        # Only support the version that ships with VS2017.
+        return not stdo.startswith(b'2.')
+    return True
+
 def detect_tests_to_run():
     # Name, subdirectory, skip condition.
     all_tests = [
@@ -473,24 +500,19 @@ def detect_tests_to_run():
         ('platform-linux', 'linuxlike', mesonlib.is_osx() or mesonlib.is_windows()),
 
         ('java', 'java', backend is not Backend.ninja or mesonlib.is_osx() or not have_java()),
-        ('C#', 'csharp', backend is not Backend.ninja or not shutil.which('mcs')),
+        ('C#', 'csharp', skip_csharp(backend)),
         ('vala', 'vala', backend is not Backend.ninja or not shutil.which('valac')),
         ('rust', 'rust', backend is not Backend.ninja or not shutil.which('rustc')),
         ('d', 'd', backend is not Backend.ninja or not have_d_compiler()),
         ('objective c', 'objc', backend not in (Backend.ninja, Backend.xcode) or mesonlib.is_windows() or not have_objc_compiler()),
+        ('objective c++', 'objcpp', backend not in (Backend.ninja, Backend.xcode) or mesonlib.is_windows() or not have_objcpp_compiler()),
         ('fortran', 'fortran', backend is not Backend.ninja or not shutil.which('gfortran')),
         ('swift', 'swift', backend not in (Backend.ninja, Backend.xcode) or not shutil.which('swiftc')),
         ('python3', 'python3', backend is not Backend.ninja),
+        ('fpga', 'fpga', shutil.which('yosys') is None),
+        ('frameworks', 'frameworks', False),
     ]
     gathered_tests = [(name, gather_tests('test cases/' + subdir), skip) for name, subdir, skip in all_tests]
-    if mesonlib.is_windows():
-        # TODO: Set BOOST_ROOT in .appveyor.yml
-        gathered_tests += [('framework', ['test cases/frameworks/1 boost'], 'BOOST_ROOT' not in os.environ)]
-    elif mesonlib.is_osx() or mesonlib.is_cygwin():
-        # Just do the BOOST test
-        gathered_tests += [('framework', ['test cases/frameworks/1 boost'], False)]
-    else:
-        gathered_tests += [('framework', gather_tests('test cases/frameworks'), False)]
     return gathered_tests
 
 def run_tests(all_tests, log_name_base, extra_args):
@@ -525,11 +547,7 @@ def _run_tests(all_tests, log_name_base, extra_args):
     # Remove this once the following issue has been resolved:
     # https://github.com/mesonbuild/meson/pull/2082
     num_workers *= 2
-    try:
-        executor = conc.ProcessPoolExecutor(max_workers=num_workers)
-    except ImportError:
-        print('Platform doesn\'t ProcessPoolExecutor, falling back to single-threaded testing\n')
-        executor = DummyExecutor()
+    executor = ProcessPoolExecutor(max_workers=num_workers)
 
     for name, test_cases, skipped in all_tests:
         current_suite = ET.SubElement(junit_root, 'testsuite', {'name': name, 'tests': str(len(test_cases))})
@@ -553,7 +571,7 @@ def _run_tests(all_tests, log_name_base, extra_args):
         for (testname, t, result) in futures:
             sys.stdout.flush()
             result = result.result()
-            if result is None or 'MESON_SKIP_TEST' in result.stdo:
+            if (result is None) or (('MESON_SKIP_TEST' in result.stdo) and (skippable(name, t))):
                 print(yellow('Skipping:'), t)
                 current_test = ET.SubElement(current_suite, 'testcase', {'name': testname,
                                                                          'classname': name})
